#include "Generic.h"
#include "Bionj.h"
#include "Dist.h"


/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

void BIONJ_Print_Mat(bmatrix *mat)
{
  int i,j;
 
  printf("    %d",mat->n_spec);  
  printf("\n");

  For(i,mat->n_spec+1)
    {
      if((i<=mat->n_spec) && (i)) printf("%s  ",mat->names[i]);
      else printf("    ");
      For(j,mat->n_spec+1)
	{
	  if(mat->dist[i][j] == -1)
	    printf("   -     ");
	  else
	    printf("%f  ",mat->dist[i][j]);
	}
      printf("\n");
    }
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

int BIONJ_K2PG(seq *data, bmatrix *mat, double g_shape, int n_otu)
{

  int i,err;
  err = 1;
  if(!(BIONJ_Compute_Dist_K2PG(data,mat,g_shape,n_otu)))
    err = 0;
  for(i=1;i<=n_otu;i++)
    {
      mat->names[i] = (char *)mCalloc(T_MAX_NAME,sizeof(char));
      strcpy(mat->names[i],data[i-1].name);
    }
  return err;
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

int BIONJ_Compute_Dist_K2PG(seq *data, bmatrix *mat, double g_shape, int n_otu)
{
  int i,j,k;
  int diff;  

  For(i,data[1].len)
    {
      For(j,n_otu)
	{
	  if(data[j].base[i] != 'A' && data[j].base[i] != 'T' && 
	     data[j].base[i] != 'G' && data[j].base[i] != 'C')
	    {
	      data[0].len--;
	      break;
	    }
	}

      if(j != n_otu) continue;

      For(j,n_otu-1)
	{
	  diff = 0;
	  for(k=j+1;k<n_otu;k++)
	    {
	      if(((data[j].base[i] == 'A' || data[j].base[i] == 'G') && 
		  (data[k].base[i] == 'C' || data[k].base[i] == 'T'))|| 
		 ((data[j].base[i] == 'C' || data[j].base[i] == 'T') && 
		  (data[k].base[i] == 'A' || data[k].base[i] == 'G'))) 
		{
		  diff++;
		  mat->Q[j+1][k+1]++;
		}
	  
	      else
		{
		  if(((data[j].base[i] == 'A' && data[k].base[i] == 'G') || 
		      (data[j].base[i] == 'G' && data[k].base[i] == 'A'))|| 
		     ((data[j].base[i] == 'C' && data[k].base[i] == 'T') || 
		      (data[j].base[i] == 'T' && data[k].base[i] == 'C'))) 
		    {
		      diff++;
		      mat->P[j+1][k+1]++;
		    }
		}
	    }
	  if(!diff) break;
	}
    }
  for(i=1;i<=n_otu;i++)
    for(j=i+1;j<=n_otu;j++)
      {
	mat->P[i][j] /= data[0].len;
	mat->Q[i][j] /= data[0].len;
	mat->P[j][i] = mat->P[i][j];
	mat->Q[j][i] = mat->Q[i][j];
	mat->dist[i][j] = (g_shape/2)*(pow(1-2*mat->P[i][j]-mat->Q[i][j],-1./g_shape) + 
				       0.5*pow(1-2*mat->Q[i][j],-1./g_shape) - 1.5);
	if(mat->dist[i][j] > 10) return 0;
	mat->dist[j][i] = mat->dist[i][j];
      }
  return 1;
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

int BIONJ_CorrectK2PG(bmatrix *mat, double g_shape)
{
  int i,j;
  int n_otu;

  n_otu = mat->n_spec;

  for(i=1;i<=n_otu-1;i++)
    {
      for(j=i+1;j<=n_otu;j++)
	{

/*  	  if(i==0 && j==1) printf("avant %d\n",&(mat->dist[i][j])); */
	  mat->dist[i][j] = (g_shape/2)*(pow(1-2*mat->P[i][j]-mat->Q[i][j],-1./g_shape) + 
					 .5*pow(1-2*mat->Q[i][j],-1./g_shape) - 1.5);
	  	  
/*  	  if(i==0 && j==1) printf("apres %d\n",&(mat->dist[i][j])); */

	  if(mat->dist[i][j] > 10) return 0;
	  mat->dist[j][i] = mat->dist[i][j];
	}
    }
  return 1;
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

bmatrix *BIONJ_Create_Mat(int n_otu)
{
  int i;
  bmatrix *mat;

  (mat) = (bmatrix *)mCalloc(1,sizeof(bmatrix));
  (mat)->n_spec = n_otu;
  (mat)->dist = (double **)mCalloc(n_otu+1,sizeof(double *));
  (mat)->P = (double **)mCalloc(n_otu+1,sizeof(double *));
  (mat)->Q = (double **)mCalloc(n_otu+1,sizeof(double *));
  (mat)->n_sons = (int **)mCalloc(n_otu+1,sizeof(int *));
  (mat)->names = (char **)mCalloc(n_otu+1,sizeof(char *));
  For(i,n_otu+1)
    {
      (mat)->dist[i] = (double *)mCalloc(n_otu+1,sizeof(double));
      (mat)->n_sons[i] = (int *)mCalloc(n_otu+1,sizeof(int));
      (mat)->P[i] = (double *)mCalloc(n_otu+1,sizeof(double));
      (mat)->Q[i] = (double *)mCalloc(n_otu+1,sizeof(double));  
      (mat)->names[i] = (char *)mCalloc(T_MAX_NAME,sizeof(char));
    }
  return mat;
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

void BIONJ_Free_Mat(bmatrix *mat)
{
  int i,n_otu;

  n_otu = mat->n_spec;  

  For(i,n_otu+1)
    {
      if(mat->names[i])
	free(mat->names[i]); 
      free(mat->P[i]);
      free(mat->Q[i]);
      free(mat->dist[i]);
      free(mat->n_sons[i]);
    }

  if(mat->names)
    free(mat->names);
  free(mat->dist);
  free(mat->n_sons);
  free(mat->P);
  free(mat->Q);
  free(mat);
}


/*;;;;;;;;;;;;;;;;;;;;;;;;;; BIONJ_Initialize        ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
; Description : This function reads an input file and return the            ;
;               delta matrix and trees: the list of taxa.                   ;
;                                                                           ;
; input       :                                                             ;
;              double **delta  : delta matrix                                ;
;              FILE *input    : pointer to input file                       ;
;              int n          : number of taxa                              ;
;              char **trees   : list of taxa                                ;
;                                                                           ;
; return value:                                                             ;
;              double **delta  : delta matrix                                ;
;              char *trees    : list of taxa                                ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/


/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

void BIONJ_Initialize(matrix *mat,double **delta, int n, POINTERS *trees)
{
    int lig;                                          /* matrix line       */
    int col;                                          /* matrix column     */
    WORD *name;

    for(lig=1; lig <= n; lig++)
    {
      name=(WORD *)mCalloc(1,sizeof(WORD));            /* taxon's name is   */
      strcpy(name->name,mat->names[lig-1]);
      name->suiv=NULL;
      trees[lig].head=name;
      trees[lig].tail=name;
      for(col= 1; col <= n; col++)
	delta[lig][col]=mat->dist[lig-1][col-1];	
    }
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

void BIONJ_Initialize_gamma(bmatrix *mat,double **delta, int n, POINTERS *trees)
{
    int lig;                                          /* matrix line       */
    int col;                                          /* matrix column     */
    WORD *name;

    for(lig=1; lig <= n; lig++)
    {
      name=(WORD *)mCalloc(1,sizeof(WORD));            /* taxon's name is   */
      strcpy(name->name,mat->names[lig]);
      name->suiv=NULL;
      trees[lig].head=name;
      trees[lig].tail=name;
      for(col= 1; col <= n; col++)
	delta[lig][col]=mat->dist[lig][col];	
    }
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

/*;;;;;;;;;;;;;;;;;;;;;;;;;;; BIONJ_Print_output;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
;                                                                           ;
; Description : This function prints out the tree in the output file.       ;
;                                                                           ;
; input       :                                                             ;
;              POINTERS *trees : pointer to the subtrees.                   ;
;              int i          : indicate the subtree i to be printed.       ;
:              FILE *output   : pointer to the output file.                 ;
;                                                                           ;
; return value: The phylogenetic tree in the output file.                   ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/



void BIONJ_Print_output(int i, POINTERS *trees,char *s_tree)
{
  WORD *parcour;
  parcour=trees[i].head;
  while(parcour != NULL)
    {
      strcat(s_tree,parcour->name);
      parcour=parcour->suiv;
    }

}


/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

char *Bionj(matrix *mat)
{
    POINTERS *trees;                        /* list of subtrees            */
    char *chain1;                           /* stringized branch-length    */
    char *chain2;                           /* idem                        */
    int *a, *b;                              /* pair to be agglomerated     */
    double **delta;                          /* delta matrix                */
    double la;                               /* first taxon's branch-length */
    double lb;                               /* second taxon's branch-length*/
    double vab;                              /* variance of Dab             */
    double lamda;
    int i;
    int ok;
    int r;                                  /* number of subtrees          */
    int n;                                  /* number of taxa              */
    int x, y;
    char *s_tree;

    
    n = mat->n_spec;


    /*   Allocation of memories    */


    s_tree=(char *)mCalloc(T_MAX_LINE,sizeof(char));
    a=(int*)mCalloc(1,sizeof(int));
    b=(int*)mCalloc(1,sizeof(int));
    chain1=(char *)mCalloc(LEN,sizeof(char));
    chain2=(char *)mCalloc(PREC,sizeof(char));
   
    

    /*      Create the delta matrix     */

    delta=(double **)mCalloc(n+1,sizeof(double*));
    for(i=1; i<= n; i++)
      delta[i]=(double *)mCalloc(n+1, sizeof(double));
      
    trees=(POINTERS *)mCalloc(n+1,sizeof(POINTERS));
 

    /*   initialise and symmetrize the running delta matrix    */

    r=n;
    *a=0;
    *b=0;
    BIONJ_Initialize(mat,delta, n, trees);
    ok=BIONJ_Symmetrize(delta, n);
    if(!ok)
      {
	printf("Problem with matrix format!\n");
	exit(-1);
      }
	
    while (r > 3)                             /* until r=3                */
      {
	BIONJ_Compute_sums_Sx(delta, n);             /* compute the sum Sx       */
	BIONJ_Best_pair(delta, r, a, b, n);          /* find the best pair by    */


	vab=BIONJ_Variance(*a, *b, delta);           /* minimizing (1)           */
	la=BIONJ_Branch_length(*a, *b, delta, r);    /* compute branch-lengths   */
	lb=BIONJ_Branch_length(*b, *a, delta, r);    /* using formula (2)        */
	if(la < BL_MIN) la = BL_MIN;
	if(lb < BL_MIN) lb = BL_MIN;
	lamda=BIONJ_Lamda(*a, *b, vab, delta, n, r); /* compute lambda* using (9)*/
	
	for(i=1; i <= n; i++)
	  {
	    if(!BIONJ_Emptied(i,delta) && (i != *a) && (i != *b))
	      {
		if(*a > i)
		  {
		    x=*a;
		    y=i;
		  }
		else
		  {
		    x=i;
		    y=*a;                         /* apply reduction formulae */
		  }                                /* 4 and 10 to delta        */
		delta[x][y]=BIONJ_Reduction4(*a, la, *b, lb, i, lamda, delta);
		delta[y][x]=BIONJ_Reduction10(*a, *b, i, lamda, vab, delta);
	     }
	  }
	strcpy(chain1,"");                     /* agglomerate the subtrees */
	strcat(chain1,"(");                    /* a and b together with the*/
	BIONJ_Concatenate(chain1, *a, trees, 0);     /* branch-lengths according */
	strcpy(chain1,"");                     /* to the NEWSWICK format   */
	strcat(chain1,":");
	sprintf(chain2, "%.5f", la);
	strcat(chain1, chain2);
	strcat(chain1,",");
	BIONJ_Concatenate(chain1,*a, trees, 1);
	trees[*a].tail->suiv=trees[*b].head;
	trees[*a].tail=trees[*b].tail;
	strcpy(chain1,"");
	strcat(chain1,":");
	sprintf(chain2, "%.5f", lb);
	strcat(chain1, chain2);
	strcat(chain1,")");
	BIONJ_Concatenate(chain1, *a, trees, 1);
	delta[*b][0]=1.0;                     /* make the b line empty     */
	trees[*b].head=NULL;
	trees[*b].tail=NULL;
	r=r-1;                                /* decrease r                */
      }
    
    
    BIONJ_Finish(delta, n, trees, s_tree);        /*compute the branch-lengths*/
    for(i=1; i<=n; i++)                     /*of the last three subtrees*/
      {                                     /*and print the tree in the */
	delta[i][0]=0.0;
	free(delta[i]);                     /* output-file              */
	trees[i].head=NULL;
	trees[i].tail=NULL;
      }

    free(delta);
    free(trees);
    free(chain2);
    free(chain1);
    free(a);
    free(b);
    
    return s_tree; 
}

/*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
;                             Utilities                                     ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/



/*;;;;;;;;;;;;;;;;;;;;;;;;;;; BIONJ_Symmetrize  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
; Description : This function verifies if the delta matrix is symmetric;    ;
;               if not the matrix is made symmetric.                        ;
;                                                                           ;
; input       :                                                             ;
;              double **delta : delta matrix                                 ;
;              int n          : number of taxa                              ;
;                                                                           ;
; return value:                                                             ;
;              int symmetric  : indicate if the matrix has been made        ;
;                               symmetric or not                            ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

int BIONJ_Symmetrize(double **delta, int n)
{
    int lig;                                         /* matrix line        */
    int col;                                         /* matrix column      */
    double value;                                     /* symmetrized value  */
    int symmetric;

    symmetric=1;
    for(lig=1; lig  <=  n; lig++)
      {
	for(col=1; col< lig; col++)
	  {
	    if(delta[lig][col] != delta[col][lig])
	      {
		value= (delta[lig][col]+delta[col][lig])/2;
		delta[lig][col]=value;
		delta[col][lig]=value;
		symmetric=0;
	      }
	  }
      }

    if(!symmetric)
      {
	printf("Problem with matrix format!\n");
	exit(-1);
      }
    return(symmetric);
}




/*;;;;;;;;;;;;;;;;;;;;;;;;;;; BIONJ_Concatenate ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
;                                                                           ;
; Description : This function concatenates a string to another.             ;
;                                                                           ;
; input       :                                                             ;
;      char *chain1    : the string to be concatenated.                     ;
;      int ind         : indicate the subtree to which concatenate the      ;
;                        string                                             ;
;      POINTERS *trees  : pointer to subtrees.                              ;
;      int post        : position to which concatenate (front (0) or        ;
;                        end (1))                                           ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

void BIONJ_Concatenate(char chain1[LEN], int ind, POINTERS *trees, int post)
{
    WORD *bran;
    
    bran=(WORD *)mCalloc(1,sizeof(WORD));
    strcpy(bran->name,chain1);
    bran->suiv=NULL;
    if(post == 0)
    {
      bran->suiv=trees[ind].head;
      trees[ind].head=bran;
    }
    else
    {
      trees[ind].tail->suiv=bran;
      trees[ind].tail=trees[ind].tail->suiv;
    }
}


/*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;BIONJ_Distance;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
; Description : This function retrieve ant return de distance between taxa  ;
;               i and j from the delta matrix.                              ;
;                                                                           ;
; input       :                                                             ;
;               int i          : taxon i                                    ;
;               int j          : taxon j                                    ;
;               double **delta : the delta matrix                            ;
;                                                                           ;
; return value:                                                             ;
;               double distance : dissimilarity between the two taxa         ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

double BIONJ_Distance(int i, int j, double **delta)
{
    if(i > j)
         return(delta[i][j]);
    else
         return(delta[j][i]);
}


/*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;BIONJ_Variance;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
; Description : This function retrieve and return the variance of the       ;
;               distance between i and j, from the delta matrix.            ;
;                                                                           ;
; input       :                                                             ;
;               int i           : taxon i                                   ;
;               int j           : taxon j                                   ;
;               double **delta  : the delta matrix                           ;
;                                                                           ;
; return value:                                                             ;
;               double distance : the variance of  Dij                       ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

double BIONJ_Variance(int i, int j, double **delta)
{
    if(i > j)
         return(delta[j][i]);
    else
         return(delta[i][j]);
}


/*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;BIONJ_Emptied ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
; Description : This function verifie if a line is emptied or not.          ;
;                                                                           ;
; input       :                                                             ;
;               int i          : subtree (or line) i                        ;
;               double **delta : the delta matrix                            ;
;                                                                           ;
; return value:                                                             ;
;               0              : if not emptied.                            ;
;               1              : if emptied.                                ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

int BIONJ_Emptied(int i, double **delta)       /* test if the ith line is emptied */
{
    return((int)delta[i][0]);
}


/*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;BIONJ_Sum_S;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
;  Description : This function retrieves the sum Sx from the diagonal       ;
;                of the delta matrix.                                       ;
;                                                                           ;
;  input       :                                                            ;
;               int i          : subtree i                                  ;
;               double **delta : the delta matrix                            ;
;                                                                           ;
;  return value:                                                            ;
;                double delta[i][i] : sum Si                                 ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

double BIONJ_Sum_S(int i, double **delta)          /* get sum Si form the diagonal */
{
    return(delta[i][i]);
}


/*;;;;;;;;;;;;;;;;;;;;;;;BIONJ_Compute_sums_Sx;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
; Description : This function computes the sums Sx and store them in the    ;
;               diagonal the delta matrix.                                  ;
;                                                                           ;
; input       :                                                             ;
;                double **delta : the delta matrix.                          ;
;                int n          : the number of taxa                        ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

void BIONJ_Compute_sums_Sx(double **delta, int n)
{
  double sum;
  int i,j;

  sum = 0.0;

  for(i= 1; i <= n ; i++)
    {if(!BIONJ_Emptied(i,delta))
      {sum=0.0;
       for(j=1; j <=n; j++)
	 {
	   if(i != j && !BIONJ_Emptied(j,delta))          /* compute the sum Si */
             sum=sum + BIONJ_Distance(i,j,delta);
	 }
      }
     delta[i][i]=sum;                            /* store the sum Si in */
    }                                           /* delta's diagonal    */
}


/*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;BIONJ_Best_pair;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
;  Description : This function finds the best pair to be agglomerated by    ;
;                minimizing the agglomerative criterion (1).                ;
;                                                                           ;
;  input       :                                                            ;
;                double **delta : the delta matrix                           ;
;                int r          : number of subtrees                        ;
;                int *a         : contain the first taxon of the pair       ;
;                int *b         : contain the second taxon of the pair      ;
;                int n          : number of taxa                            ;
;                                                                           ;
;  return value:                                                            ;
;                int *a         : the first taxon of the pair               ;
;                int *b         : the second taxon of the pair              ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

void BIONJ_Best_pair(double **delta, int r, int *a, int *b, int n)
{
    double Qxy;                         /* value of the criterion calculated*/
    int x,y;                            /* the pair which is tested         */
    double Qmin;                        /* current minimun of the criterion */
    double Qold;

    Qmin=1.e30;
    for(x=1; x <= n; x++)
      {
        if(!BIONJ_Emptied(x,delta))
	  {
	    for(y=1; y < x; y++)
	      {
		if(!BIONJ_Emptied(y,delta))
		  {
                    Qxy=BIONJ_Agglomerative_criterion(x,y,delta,r,n);
		    
		    if((Qxy < Qmin))
		      {
			Qold=Qmin;
			Qmin=Qxy;
			*a=x;          
			*b=y;		       
		      }
		  }  
	      }
	  }
      }

}


/*;;;;;;;;;;;;;;;;;;;;;;BIONJ_Finish_branch_length;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
;  Description :  Compute the length of the branch attached                 ;
;                 to the subtree i, during the final step                   ;
;                                                                           ;
;  input       :                                                            ;
;                int i          : position of subtree i                     ;
;                int j          : position of subtree j                     ;
;                int k          : position of subtree k                     ;
;                double **delta :                                            ;
;                                                                           ;
;  return value:                                                            ;
;                double length  : The length of the branch                   ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

double BIONJ_Finish_branch_length(int i, int j, int k, double **delta)
{
    double length;
    length=0.5*(BIONJ_Distance(i,j,delta) + BIONJ_Distance(i,k,delta)
                -BIONJ_Distance(j,k,delta));
    return(length);
}


/*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;BIONJ_Finish;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;
;  Description : This function compute the length of the lasts three        ;
;                subtrees and write the tree in the output file.            ;
;                                                                           ;
;  input       :                                                            ;
;                double **delta  : the delta matrix                          ;
;                int n           : the number of taxa                       ;
;                WORD *trees   : list of subtrees                           ;
;                                                                           ;
;                                                                           ;
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/

void BIONJ_Finish(double **delta, int n, POINTERS *trees, char *s_tree)
{
    int l=1;
    int i=0;
    double length;
    char *str;
    WORD *bidon;
    WORD *ele;
    int last[3];                        /* the last three subtrees     */

    str=(char *)mCalloc(LEN,sizeof(char));

 
    while(l <= n)
      {                                       /* find the last tree subtree  */
	if(!BIONJ_Emptied(l, delta))
	  {last[i]=l;
	   i++;
	  }
	l++;
      }

    length=BIONJ_Finish_branch_length(last[0],last[1],last[2],delta);
    if(length < BL_MIN) length = BL_MIN;
    strcat(s_tree,"(");
    BIONJ_Print_output(last[0],trees,s_tree);
    strcat(s_tree,":");
    sprintf(str, "%.5f", length);
    strcat(s_tree,str);
    strcat(s_tree,",");

    length=BIONJ_Finish_branch_length(last[1],last[0],last[2],delta);
    if(length < BL_MIN) length = BL_MIN;
    BIONJ_Print_output(last[1],trees,s_tree);
    strcat(s_tree,":");
    sprintf(str, "%.5f", length);
    strcat(s_tree,str);
    strcat(s_tree,",");

    length=BIONJ_Finish_branch_length(last[2],last[1],last[0],delta);
    if(length < BL_MIN) length = BL_MIN;
    BIONJ_Print_output(last[2],trees,s_tree);
    strcat(s_tree,":");
    sprintf(str, "%.5f", length);
    strcat(s_tree,str);
    strcat(s_tree,");\n");

    for(i=0; i < 3; i++)
    {
       bidon=trees[last[i]].head;
       ele=bidon;
       while(bidon!=NULL)
       {
          ele=ele->suiv;
          free(bidon);
          bidon=ele;
        }
    }

    free(str);
}


/*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*\
;                                                                           ;   
;                          Formulae                                         ;
;                                                                           ;  
\*;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;*/


double BIONJ_Agglomerative_criterion(int x, int y, double **delta, int r, int n_otu)
{
  double Qxy;

  Qxy=(r-2)*BIONJ_Distance(x,y,delta)                           /* Formula (1) */
        -BIONJ_Sum_S(x,delta)
        -BIONJ_Sum_S(y,delta); 


  return(Qxy);                       
}


/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

double BIONJ_Branch_length(int a, int b, double **delta, int r)
{
    double length;
    length=0.5*(BIONJ_Distance(a,b,delta)                         /* Formula (2) */
           +(BIONJ_Sum_S(a,delta)
           -BIONJ_Sum_S(b,delta))/(r-2)); 
    return(length);                                   
}


/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

double BIONJ_Reduction4(int a, double la, int b, double lb, int i, double lamda,
                 double **delta)
{
  double Dui;
  Dui=lamda*(BIONJ_Distance(a,i,delta)-la)
     +(1-lamda)*(BIONJ_Distance(b,i,delta)-lb);                  /* Formula (4) */
  return(Dui);
}


/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

double BIONJ_Reduction10(int a, int b, int i, double lamda, double vab,
                  double **delta)
{
  double Vci;
  Vci=lamda*BIONJ_Variance(a,i,delta)+(1-lamda)*BIONJ_Variance(b,i,delta)
     -lamda*(1-lamda)*vab;                                    /*Formula (10)  */
  return(Vci);
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

double BIONJ_Lamda(int a, int b, double vab, double **delta, int n, int r)
{
    double lamda=0.0;
    int i;
    
    if(vab==0.0)
      lamda=0.5;
    else
      {
	for(i=1; i <= n ; i++)
	  {
	    if(a != i && b != i && !BIONJ_Emptied(i,delta))
	      lamda=lamda + (BIONJ_Variance(b,i,delta) - BIONJ_Variance(a,i,delta));
	  }
	lamda=0.5 + lamda/(2*(r-2)*vab);             
      }                                               /* Formula (9) and the  */
    if(lamda > 1.0)                            /* constraint that lamda*/
      lamda = 1.0;                              /* belongs to [0,1]     */
    if(lamda < 0.0)
      lamda=0.0;
    return(lamda);
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

void BIONJ_Best_pair_gamma(bmatrix **mat, int r, int *a, int *b, int n, int nalpha)
{
    double **Qxy;                         /* value of the criterion calculated*/
    int x,y;                            /* the pair which is tested         */
    double Qmin;                        /* current minimun of the criterion */
    double erold,ernew;
    int nalive;
    int i,j,iold;
    int **tabx,**taby;
    int xmin,ymin,xminold,yminold;

    Qxy = (double **)mCalloc(nalpha,sizeof(double *));
    For(i,nalpha) Qxy[i] = (double *)mCalloc(n*n,sizeof(double));
    tabx = (int **)mCalloc(nalpha,sizeof(int *));
    For(i,nalpha) tabx[i] = (int *)mCalloc(n*n,sizeof(int));
    taby = (int **)mCalloc(nalpha,sizeof(int *));
    For(i,nalpha) taby[i] = (int *)mCalloc(n*n,sizeof(int));

    iold = xmin = ymin = xminold = yminold = 0;
    For(i,nalpha)
      {
	nalive=0;
	for(x=1;x<= n;x++)
	  {
	    if(!BIONJ_Emptied(x,mat[nalpha-1]->dist))
	      {
		for(y=1;y<x;y++)
		  {
		    if(!BIONJ_Emptied(y,mat[nalpha-1]->dist))
		      {
			Qxy[i][nalive]=BIONJ_Agglomerative_criterion(x,y,mat[i]->dist,r,n);
			tabx[i][nalive]=x;
			taby[i][nalive]=y;
			nalive++;
		      }
		  }
	      }
	  }

	Qmin = 1.e10;
	For(j,nalive)
	  if(Qxy[i][j] < Qmin)
	    {
	      
	      Qmin = Qxy[i][j];
	      xmin = tabx[i][j];
	      ymin = taby[i][j];
	      
	    }
/*  	fprintf(stderr,"%d %d\n",xmin,ymin); */
	if(!i)
	  {
	    xminold = xmin;
	    yminold = ymin;
	    iold = 0;
	    *a = xmin;
	    *b = ymin;
	  }
	if((xmin != xminold) || (ymin != yminold))
	  {
	    erold = BIONJ_Ecart_Relatif_Min(Qxy[iold],nalive);
	    ernew = BIONJ_Ecart_Relatif_Min(Qxy[i],nalive);
	    if(ernew > erold)
	      {
		*a = xmin;
		*b = ymin;
		xminold = xmin;
		yminold = ymin;
		iold = i;
	      }
	  }
      }
    For(i,nalpha){free(tabx[i]);free(taby[i]);free(Qxy[i]);}
    free(tabx);free(taby);free(Qxy);
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

double BIONJ_Ecart_Relatif_Min(double *tab, int nelem)
{
  int i,j;
  double min,er;
  double *dmoy;

  dmoy = (double *)mCalloc(nelem,sizeof(double));

  er=0.0;
  min = +1.e10;
  For(i,nelem)
    {
      if(tab[i] < min) min = tab[i];
      for(j=0;j<nelem;j++)
	{
	  if(j != i)
	    dmoy[i] += fabs(tab[i] - tab[j]);
	}
    }
  For(i,nelem)
    {
      dmoy[i]/=nelem;
      er += fabs(tab[i] - min)/dmoy[i];
    }
  free(dmoy);
  return er/nelem;
}
  
/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

char *BIONJ_Seq(seq *data, int n_otu)
{
    POINTERS *trees;                        /* list of subtrees            */
    char *chain1;                           /* stringized branch-length    */
    char *chain2;                           /* idem                        */
    int *a, *b;                              /* pair to be agglomerated     */
    double **delta;                          /* delta matrix                */
    double *la;                               /* first taxon's branch-length */
    double *lb;                               /* second taxon's branch-length*/
    double *vab;                              /* variance of Dab             */
    double *lamda;
    int i,j;
    int r;                                  /* number of subtrees          */
    int n;                                  /* number of taxa              */
    int x, y;
    char *s_tree;
    bmatrix **mat;
    double *alpha;
    int nalpha;



    n = n_otu;
    nalpha = 10;
    
    /*   Allocation of memories    */

    la = (double *)mCalloc(nalpha,sizeof(double));
    lb = (double *)mCalloc(nalpha,sizeof(double));
    vab = (double *)mCalloc(nalpha,sizeof(double));
    lamda = (double *)mCalloc(nalpha,sizeof(double));
    s_tree=(char *)mCalloc(T_MAX_LINE,sizeof(char));
    a=(int*)mCalloc(1,sizeof(int));
    b=(int*)mCalloc(1,sizeof(int));
    chain1=(char *)mCalloc(LEN,sizeof(char));
    chain2=(char *)mCalloc(PREC,sizeof(char));
    alpha=(double *)mCalloc(nalpha,sizeof(double));
    mat = (bmatrix **)mCalloc(nalpha,sizeof(bmatrix *));

    alpha[0]=0.2;
    alpha[1]=0.3;
    alpha[2]=0.4;
    alpha[3]=0.5;
    alpha[4]=0.6;
    alpha[5]=0.7;
    alpha[6]=0.8;
    alpha[7]=0.9;
    alpha[8]=1;
    alpha[9]=1000;


    For(i,nalpha) 
      {
	mat[i] = (bmatrix *)mCalloc(1,sizeof(bmatrix));
	mat[i] = (bmatrix *)BIONJ_Create_Mat(n_otu);
     }
    For(i,nalpha)  if(!BIONJ_K2PG(data,mat[i],alpha[i],n_otu)) exit(1);
      

    /*      Create the delta matrix     */

    delta=(double **)mCalloc(n+1,sizeof(double*));
    for(i=1; i<= n; i++)
      delta[i]=(double *)mCalloc(n+1, sizeof(double));
      
    trees=(POINTERS *)mCalloc(n+1,sizeof(POINTERS));
 


    r=n;
    *a=0;
    *b=0;
    BIONJ_Initialize_gamma(mat[0],delta,n,trees);
	

    while (r > 3)                             /* until r=3                */
      {
	For(j,nalpha) BIONJ_Compute_sums_Sx(mat[j]->dist, n);             
	BIONJ_Best_pair_gamma(mat,r,a,b,n,nalpha);

	For(j,nalpha) vab[j]= BIONJ_Variance(*a,*b,mat[j]->dist);
	For(j,nalpha) la[j] = BIONJ_Branch_length(*a,*b,mat[j]->dist,r);
	For(j,nalpha) lb[j] = BIONJ_Branch_length(*b,*a,mat[j]->dist,r); 
	For(j,nalpha) lamda[j] = BIONJ_Lamda(*a,*b,vab[j],mat[j]->dist,n,r);

	for(i=1;i<=n;i++)
	  {
	    if(!BIONJ_Emptied(i,mat[0]->dist) && (i != *a) && (i != *b))
	      {
		if(*a > i){x=*a;y=i;}
		else
		  {x=i;y=*a;}

		For(j,nalpha)
		  {
		    mat[j]->dist[x][y]=BIONJ_Reduction4(*a,la[j],*b,lb[j],i,lamda[j],mat[j]->dist);
		    mat[j]->dist[y][x]=BIONJ_Reduction10(*a,*b,i,lamda[j],vab[j],mat[j]->dist);
		  }
	      }
	  }

	strcpy(chain1,"");                     /* agglomerate the subtrees */
	strcat(chain1,"(");                    /* a and b together with the*/
	BIONJ_Concatenate(chain1, *a, trees, 0);     /* branch-lengths according */
	strcpy(chain1,"");                     /* to the NEWSWICK format   */
	strcat(chain1,":");
	sprintf(chain2, "%.5f", la[0]);
	strcat(chain1, chain2);
	strcat(chain1,",");
	BIONJ_Concatenate(chain1,*a, trees, 1);
	trees[*a].tail->suiv=trees[*b].head;
	trees[*a].tail=trees[*b].tail;
	strcpy(chain1,"");
	strcat(chain1,":");
	sprintf(chain2, "%.5f", lb[0]);
	strcat(chain1, chain2);
	strcat(chain1,")");
	BIONJ_Concatenate(chain1, *a, trees, 1);
	For(j,nalpha) mat[j]->dist[(*b)][0]=1.0;
	trees[*b].head=NULL;
	trees[*b].tail=NULL;
	r=r-1;                                /* decrease r                */
      }
    BIONJ_Finish(mat[0]->dist,n,trees,s_tree);        /*compute the branch-lengths*/

    for(i=1; i<=n; i++)                     /*of the last three subtrees*/
      {                                     /*and print the tree in the */
	delta[i][0]=0.0;
	free(delta[i]);                     /* output-file              */
	trees[i].head=NULL;
	trees[i].tail=NULL;
      }

    free(la);
    free(lb);
    free(lamda);
    free(delta);
    free(alpha);
    free(trees);
    free(chain2);
    free(chain1);
    free(a);
    free(b);
    For(i,nalpha)
      BIONJ_Free_Mat(mat[i]);

    return s_tree; 
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/




