#include "Generic.h"
#include "Dist.h"

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/


int K2PG(seq *data, matrix *mat, double g_shape, int n_otu)
{
  int i;
  int err;

  err = 1;

  if(!(Compute_Dist_K2PG(data,mat,g_shape,n_otu)))
    err = 0;

  For(i,n_otu)
    {
      strcpy(mat->names[i],data[i].name);
    }
  return err;
}


/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/


int JCG(seq *data, matrix *mat, double g_shape, int n_otu)
{
  int i;
  int err;

  err = 1;

  if(!(Compute_Dist_JCG(data,mat,g_shape,n_otu)))
    err = 0;


  For(i,n_otu)
    {
      mat->names[i] = (char *)mCalloc(T_MAX_NAME,sizeof(char));
      strcpy(mat->names[i],data[i].name);
    }
  
  return err;
}


/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/


int Compute_Dist_K2PG(seq *data, matrix *mat, double g_shape, int n_otu)
{
  int i,j,k;
  int diff;  

  For(i,data[1].len)
    {
      For(j,n_otu)
	{
	  if(data[j].base[i] != 'A' && data[j].base[i] != 'T' && 
	     data[j].base[i] != 'G' && data[j].base[i] != 'C' && data[j].base[i] != 'U')
	    {
	      data[0].len--;
	      break;
	    }
	}

      if(j != n_otu) continue;

      For(j,n_otu-1)
	{
	  diff = 0;
	  for(k=j+1;k<n_otu;k++)
	    {
	      if(((data[j].base[i] == 'A' || data[j].base[i] == 'G') && 
		  (data[k].base[i] == 'C' || (data[k].base[i] == 'T' || data[k].base[i] == 'U')))|| 
		 ((data[j].base[i] == 'C' || (data[j].base[i] == 'T' || data[k].base[i] == 'U')) && 
		  (data[k].base[i] == 'A' || data[k].base[i] == 'G'))) 
		{
		  diff++;
		  mat->Q[j][k]++;
		}
	  
	      else
		{
		  if( (data[j].base[i] == 'A' && data[k].base[i] == 'G') || 
		      (data[j].base[i] == 'G' && data[k].base[i] == 'A') || 
		      (data[j].base[i] == 'C' && (data[k].base[i] == 'T' || data[k].base[i] == 'U')) || 
		      ((data[j].base[i] == 'T' || data[k].base[i] == 'U') && data[k].base[i] == 'C'))
		    {
		      diff++;
		      mat->P[j][k]++;
		    }
		}
	    }
	  if(!diff) break;
	}
    }


 
  For(i,n_otu-1)
    for(j=i+1;j<n_otu;j++)
      {
/*  	if(!(mat->P[i][j]))  */
/*  	  mat->P[i][j] = 1; */
/*  	if(!(mat->Q[i][j]))  */
/*  	  mat->Q[i][j] = 1; */
	mat->P[i][j] /= data[0].len;
	mat->Q[i][j] /= data[0].len;
	mat->P[j][i] = mat->P[i][j];
	mat->Q[j][i] = mat->Q[i][j];
	if((1-2*mat->P[i][j]-mat->Q[i][j] < .0) || (1-2*mat->Q[i][j] < .0)) return 0;
	mat->dist[i][j] = (g_shape/2)*(pow(1-2*mat->P[i][j]-mat->Q[i][j],-1./g_shape) + 0.5*pow(1-2*mat->Q[i][j],-1./g_shape) - 1.5);
	if((mat->dist[i][j] > 10) || (mat->dist[i][j] < 0)) return 0;
	mat->dist[j][i] = mat->dist[i][j];
      }

  return 1;

}


/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/


int Compute_Dist_JCG(seq *data, matrix *mat, double g_shape, int n_otu)
{
  
  int i,j,k;
  int diff;


  For(i,data[1].len)
    {
      For(j,n_otu)
	{
	  if(data[j].base[i] != 'A' && data[j].base[i] != 'T' && 
	     data[j].base[i] != 'G' && data[j].base[i] != 'C')
	    {
	      data[0].len--;
	      break;
	    }
	}

      if(j != n_otu) continue;


      For(j,n_otu-1)
	{
	  diff = 0;
	  for(k=j+1;k<n_otu;k++)
	    {
	      if(data[j].base[i] != data[k].base[i]) 
		{
		  diff++;
		  mat->P[j][k]++;
		}
	    }
	  if(!diff) break;
	}
    }


  For(i,n_otu-1)
    for(j=i+1;j<n_otu;j++)
      {
/*  	if(! mat->P[i][j]) mat->P[i][j] = 1; */
	mat->P[i][j] /= data[0].len;
	if((1-(4./3.)*mat->P[i][j] < .0)) return 0;
	mat->P[j][i] = mat->P[i][j];
	if(mat->P[i][j] > 3./4) mat->P[i][j] = 3./4;
	mat->dist[i][j] = (3./4) * g_shape * (pow(1 - (4./3)*mat->P[i][j],-1./g_shape) - 1);
	if((mat->dist[i][j] > 10)) return 0;
	mat->dist[j][i] = mat->dist[i][j];
      }

  return 1;
}


/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/


int CorrectK2PG(matrix *mat, double g_shape)
{

  int i,j;
  int n_otu;

  n_otu = mat->n_spec;


  For(i,n_otu-1)
    {
      for(j=i+1;j<n_otu;j++)
	{

/*  	  if(i==0 && j==1) printf("avant %d\n",&(mat->dist[i][j])); */
	  mat->dist[i][j] = (g_shape/2)*(pow(1-2*mat->P[i][j]-mat->Q[i][j],-1./g_shape) + .5*pow(1-2*mat->Q[i][j],-1./g_shape) - 1.5);
	  	  
/*  	  if(i==0 && j==1) printf("apres %d\n",&(mat->dist[i][j])); */

	  if(mat->dist[i][j] > 10) return 0;
	  mat->dist[j][i] = mat->dist[i][j];
	}
    }

  return 1;

}



/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/


int CorrectJCG(matrix *mat, double g_shape)
{

  int i,j;
  int n_otu;

  n_otu = mat->n_spec;


  For(i,n_otu-1)
    {
      for(j=i+1;j<n_otu;j++)
	{
	  mat->dist[i][j] = (3./4) * g_shape * (pow(1 - (4./3)*mat->P[i][j],-1./g_shape) - 1);
	  if(mat->dist[i][j] > 10) return 0;
	  mat->dist[j][i] = mat->dist[i][j];
	}
    }
  
  return 1;

}


/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

