
#ifndef GENERIC_H
#define GENERIC_H

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>


#define T_MAX_FILE  100
#define T_MAX_LINE  100000
#define T_MAX_NAME  200
#define T_MAX_SEQ   10000
#define T_MAX_ALPHA 500

#define For(i,n) for(i=0; i<n; i++)




/*********************************************************/

typedef struct __Branche {
  struct __Node *boutg, *boutd;
  double l;
  double depth;
  struct __Node **bipart;
  int n_bipart;
  int common;
}branch;

/*********************************************************/

typedef struct __Matrix {
  char **names;
  double **dist;
  double **P;
  double **Q;
  int **n_sons;
  int n_spec;
}matrix;


/*********************************************************/

typedef struct __Node {
  struct __Node **v;
  struct __Node *arbre_gauche, *arbre_droite;
  struct __Branche **b;
  int num_arbre_gauche, num_arbre_droite, num_arbre_haut;
  double *l;              /* longueurs des branches emmergentes */
  int *n_sons;            /* n_sons[0] : nombre de feuilles vers v[0] ... */
  int **sons;             /* numeros des feuilles descendantes vers v[0],v[1] et v[2] */
  int num;                /* si num < n_otu  -> noeud peripherique */
  char *name;             /* nom de la feuille (< 40 caracteres) */
  double *additiv;
  double *Wadditiv;
  int best_add[3];
}node;


/*********************************************************/


typedef struct __Arbre {
  struct __Node **noeuds;         /* liste des noeuds de l'arbre */
  int n_otu;             /* nombre de feuilles */
  struct __Branche **tree_branch;  
  int n_branch;
  int n_branch_neg;
}arbre;



/*********************************************************/


typedef struct __Seq {
  char *name;
  int len;
  char *base;
}seq;

/*********************************************************/

typedef struct __Options {
  int format;
  int mult;
  int each;
  int ocrit;
  int tree_method;
  int otree;
  int model;
  int interleaved;
  int fixed;
  double f_alpha;
  FILE *in_alpha;
}option;

/*********************************************************/

/* BIONJ et UNJ */

#define PREC 8    /* precision of branch-lengths  */
#define LEN  100       /* length of taxon names  */
                                
 
/*********************************************************/


typedef struct __word
{
    char name[LEN];
    struct __word *suiv;
}WORD;


/*********************************************************/


typedef struct __pointers
{
  WORD *head;
  WORD *tail;
}POINTERS;


void Echange_String(char **a, char **b);
void Echange_Vect_Double(double **a, double **b);
void Echange_Elem_Double(double *a, double *b);
void *mCalloc(int nb, size_t taille);
void *mRealloc(void *p, int nb, size_t taille);
char **Sub_Trees(char *tree);
arbre *Read_Tree(char *s_tree);
void R_rtree(char *s_tree, node *pere, node **noeud, int n_otu, int *n_int, int *n_ext);
void Init_Node(node *noeud_func);
int Next_Par(char *s, int pos);
void Test_Tree(char *s);
void M_Sons(node **noeud, int n_otu);
void R_M_Sons(node *pere, node *fils, int n_otu);
void M_Sons_Ext(node *ext, int n_otu);
int Tri_String(const void *a, const void *b);
int Tri_Int_Decrois(const void *a, const void *b);
int Tri_Double_Decrois(const void *a, const void *b);
int Tri_Int_Crois(const void *a, const void *b);
int Tri_Double_Crois(const void *a, const void *b);
void Free_Tree(arbre *tree);
void Free_Branch(branch *edge);
void Free_Noeud(node *noeud);
void Init_Tree(arbre *tree_func, node **noeud_func);
matrix *Read_Mat(FILE *in);
void M_Order(node **noeud, matrix *mat);
void Print_Mat(matrix *mat);
void Free_Seq(seq *Data, int n_spec);
void Free_Mat(matrix *mat);
double TreeBl(arbre *tree);
void R_TreeBl(node *pere, node *fils, double *T, int n_otu);
void Init_Mat(matrix *mat);
matrix *Create_Mat(int n_otu);
double **Lbranch2Dist(arbre *tree);
void R_m_dist(node *pere, node *fils, double **dist, int n_otu);
char *Write_Tree(arbre *tree);
void R_wtree(node *pere,node *fils,char *s_tree,int n_otu);
arbre *Copy_Tree(arbre *original);
void R_tree_branch(node *pere,node *fils, arbre *tree);
void Norm_Mat(matrix *mat, double normeur);
void Print_Seq(seq *data, int n_otu);
int Pos_Min(double *tab, int n_elem);
seq *Read_Seq_Interleaved(FILE *in, int *n_otu);
seq *Read_Seq_Sequential(FILE *in, int *n_otu);
void Uppercase(char *ch);
int Read_One_Line_Seq(seq *data, int num_otu, int *position, FILE *in);

#endif











