#include "Generic.h"
#include "Main.h"
#include "Additiv.h"
#include "Dist.h"
#include "Bionj.h"
#include "Read_In.h"


int main(int argc, char **argv)
{
  FILE *in;
  option *opt;
  char *file;
 
  file = (char *)mCalloc(T_MAX_FILE,sizeof(char));

  


  printf("\n  Data set file name : ");	fflush(stdout);	/* ++pac */
  scanf("%s",file);
  while(!(in = fopen(file,"r")))
    {
      printf("\n  Can't open file %s, enter a new name : ",file);	fflush(stdout);	/* ++pac */
      scanf("%s",file);
    }  
  free(file);

  opt = Get_Options();

  if(!opt->format) From_Sequences(opt,&in);
  else From_Distances(opt,&in);
  
  return 1;
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

double Criterion_Sequences(matrix *mat, double *alpha, option *opt, double *q_val)
{
  int current, check;
  double *additiv,a_val;
  char *s_tree;
  arbre *tree;
 

  current = 0;
  while(alpha[current] != -1) current++;
  if(!current) exit(1);

  additiv = (double *)mCalloc(current,sizeof(double));
  
  current = 0;
  check = 1;
  while(alpha[current] > .0)
    {  
      if(!(opt->model)) check =  CorrectJCG(mat,alpha[current]);
      else if(opt->model == 1) check =  CorrectK2PG(mat,alpha[current]);

      if(!check)
	{
	  additiv[current] = -1.0;
	  current++;
	  continue;
	}      

      s_tree = (char *)Bionj(mat);
      tree = (arbre *)Read_Tree(s_tree);
      Norm_Mat(mat,1);
      additiv[current] = Q_Add_Topo(tree,mat);
      if((!opt->mult) && (!opt->fixed)) 
	printf(" %3d alpha = %10.3f; Q-val = %10E \n",
	       current+1,
	       alpha[current],
	       additiv[current]);
      free(s_tree);
      Free_Tree(tree); free(tree);
      current++;
    }

  a_val= Pos_Min(additiv,current);
  *q_val = additiv[(int)a_val];
  a_val = (double)alpha[(int)a_val];
  free(additiv);
  return a_val;
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

double Criterion_Distances(FILE **in_mat, double *alpha, double *q_val)
{
  matrix *mat;
  int current;
  double *additiv,a_val;
  char *s_tree;
  arbre *tree;
 

  current = 0;
  while(alpha[current] != -1) current++;
  if(!current) exit(1);

  additiv = (double *)mCalloc(current,sizeof(double));
  
  current = 0;
  while((mat = Read_Mat(*in_mat)))
    {  
      s_tree = (char *)Bionj(mat);
      tree = (arbre *)Read_Tree(s_tree);
      Norm_Mat(mat,1);
      additiv[current] = Q_Add_Topo(tree,mat);
      free(s_tree);
      Free_Tree(tree); free(tree);
      Free_Mat(mat);
      current++;
      printf(" Matrix %3d done (alpha = %8.3f; Q-val = %10E) \n",
	     current,
	     alpha[current-1],
	     additiv[current-1]);
    }
  
  a_val= Pos_Min(additiv,current);
  *q_val = additiv[(int)a_val];
  a_val = (double)alpha[(int)a_val];
  free(additiv);
  return a_val;
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

void From_Sequences(option *opt, FILE **in_seq)
{
  seq *data;
  matrix *mat;
  double *alpha;
  double alpha_min,first_alpha_min,qval;
  int n_otu,first,cpt;
  FILE *out_Tree, *out_Qval;
  char *s;

  out_Tree = out_Qval = NULL;

  out_Tree = fopen("out_Tree","w");
  out_Qval = fopen("out_Qval","w");
  

  if(!opt->fixed) alpha = Get_Alpha(opt);
  else
    {
      alpha = (double *)mCalloc(2,sizeof(double));
      alpha[0] = opt->f_alpha;
      alpha[1] = -1.;
    }


  if(!(data = Get_Seq((*in_seq),&n_otu,opt))) 
    {
      printf("Problem during sequence reading\n");
      exit(-1);
    }

  if(!(mat = Create_Mat(n_otu))) 
    {
      printf("Can't create the distance matrix with this data set\n");
      exit(-1);
    }

  if(!(K2PG(data,mat,10000,n_otu)))
    {
      printf("Can't compute distances with this data set\n");
      exit(-1);
    }

  first_alpha_min = .0;
  first = cpt = 1;

  while(data)
    {
      alpha_min = Criterion_Sequences(mat,alpha,opt,&qval);
      
      if(first) first_alpha_min = alpha_min;
      
      printf(" Data set %4d checked (alpha* = %8.3f ; Q-val = %10E) \n",
	     cpt++,
	     alpha_min,
	     qval);

      if(!opt->fixed)
	fprintf(out_Qval,"alpha* = %8.3f\tQ-val = %10E\n",alpha_min,qval);
      else
	fprintf(out_Qval,"used defined alpha  = %8.3f\tQ-val = %10E\n",alpha_min,qval);
	
	
      if(opt->each)
	{	  
	  if(!(opt->model)) CorrectJCG(mat,alpha_min);
	  else if(opt->model == 1) CorrectK2PG(mat,alpha_min);  
	  fprintf(out_Tree,"%s",s=(char *)Bionj(mat));
	}
      else
	{
	  if(first)
	    {
	      free(alpha);
	      alpha = (double *)mCalloc(2,sizeof(double));
	      alpha[0] = first_alpha_min;
	      alpha[1] = -1.;
	    }
	  if(!(opt->model)) CorrectJCG(mat,first_alpha_min);
	  else if(opt->model == 1) CorrectK2PG(mat,first_alpha_min);  
	  fprintf(out_Tree,"%s",s=(char *)Bionj(mat));
	}
      
      free(s);
      first = 0;

      Free_Seq(data,n_otu);
      if(!(opt->mult)) break;
      if(!(data = Get_Seq((*in_seq),&n_otu,opt))) break;
      K2PG(data,mat,10000,n_otu);
      fflush(NULL);
    }
  fprintf(out_Qval,"Program ended succefully\n");
  free(alpha);
  fclose(out_Tree);
  fclose(out_Qval);
  fclose(*in_seq);
}

/*:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:!:*/

void From_Distances(option *opt, FILE **in_dist)
{
  double qval;
  FILE *out_Tree, *out_Qval;
  double *alpha,alpha_min;

  out_Tree = out_Qval = NULL;

  alpha = Get_Alpha(opt);
  out_Tree = fopen("out_Tree","w");
  out_Qval = fopen("out_Qval","w");
  alpha_min = Criterion_Distances(in_dist,alpha,&qval);
  printf("alpha* = %8.3f ; Q-val = %10E) \n",alpha_min,qval);
  fprintf(out_Qval,"alpha = %8.3f\tQ-val = %10E\n",alpha_min,qval);
  fprintf(out_Qval,"Program ended succefully\n");
  free(alpha);
  fclose(out_Tree);
  fclose(out_Qval);
  fclose(*in_dist);
}
